"""Defines a manifest loader for STACS.

SPDX-License-Identifier: BSD-3-Clause
"""

#
# TODO: Implement the manifest loader. This should take the contents of a manifest
#       which matches the stacs.model.manifest.Format schema. It should also check
#       whether all requested files exist, and generate MD5 sums for them if not
#       specified in the manifest 'Entry' (stacs.model.manifest.Entry).
#
