"""Defines loaders used by STACS.

SPDX-License-Identifier: BSD-3-Clause
"""

from stacs.loader import archive  # noqa:F401
from stacs.loader import filepath  # noqa:F401
from stacs.loader import manifest  # noqa:F401
