"""STACS Setup.

SPDX-License-Identifier: BSD-3-Clause
"""

from setuptools import setup

# Defer to setupmeta.
setup(
    name="stacs",
    setup_requires="setupmeta",
    description="Static Token And Credential Scanner.",
)
