"""STACS - Static Token And Credential Scanner.

SPDX-License-Identifier: BSD-3-Clause
"""

__title__ = "stacs"
__summary__ = "Static Token And Credential Scanner."
__version__ = "0.1.1"
__author__ = "Peter Adkins"
__uri__ = "https://www.github.com/stacscan/stacs/"
__license__ = "BSD-3-Clause"
